@@ -28,6 +28,11 @@
 
 exception Canceled
 
+module Int_map = Map.Make(struct type t = int let compare = compare end)
+
+type data = (unit -> unit) Int_map.t
+  (* Type of data carried by threads *)
+
 type +'a t
 type -'a u
 
@@ -43,7 +48,10 @@ type 'a thread_state =
   | Repr of 'a thread_repr
       (* [Repr t] a thread which behaves the same as [t] *)
 
-and 'a thread_repr = 'a thread_state ref
+and 'a thread_repr = {
+  mutable state : 'a thread_state;
+  (* The state of the thread *)
+}
 
 and 'a sleeper = {
   cancel : (unit -> unit) ref;
@@ -72,14 +80,41 @@ external wakener_repr : 'a u -> 'a thread_repr = "%identity"
    being cleaned: *)
 let max_removed = 42
 
+(* +-----------------------------------------------------------------+
+   | Local storage                                                   |
+   +-----------------------------------------------------------------+ *)
+
+type 'a key = {
+  id : int;
+  mutable store : 'a option;
+}
+
+let next_key_id = ref 0
+
+let new_key () =
+  let id = !next_key_id in
+  next_key_id := id + 1;
+  { id = id; store = None }
+
+let current_data = ref Int_map.empty
+
+let get key =
+  try
+    Int_map.find key.id !current_data ();
+    let value = key.store in
+    key.store <- None;
+    value
+  with Not_found ->
+    None
+
 (* +-----------------------------------------------------------------+
    | Restarting/connecting threads                                   |
    +-----------------------------------------------------------------+ *)
 
 (* Returns the representative of a thread, updating non-direct references: *)
 let rec repr_rec t =
-  match !t with
-    | Repr t' -> let t'' = repr_rec t' in if t'' != t' then t := Repr t''; t''
+  match t.state with
+    | Repr t' -> let t'' = repr_rec t' in if t'' != t' then t.state <- Repr t''; t''
     | _       -> t
 let repr t = repr_rec (thread_repr t)
 
@@ -108,44 +143,53 @@ let rec run_waiters_rec state ws rem =
 
 (* Run all waiters waiting on [t]: *)
 let run_waiters waiters state =
-  run_waiters_rec state waiters []
+  let save = !current_data in
+  run_waiters_rec state waiters [];
+  current_data := save
 
-(* Restarts a sleeping thread [t]:
-
-   - run all its waiters
-   - set his state to the terminated state [state]
-*)
-let restart t state caller =
+let wakeup t v =
   let t = repr_rec (wakener_repr t) in
-  match !t with
+  match t.state with
     | Sleep{ waiters = waiters } ->
-        t := state;
+        let state = Return v in
+        t.state <- state;
         run_waiters waiters state
     | Fail Canceled ->
         (* Do not fail if the thread has been canceled: *)
         ()
     | _ ->
-        invalid_arg caller
+        invalid_arg "Lwt.wakeup"
+
+let wakeup_exn t e =
+  let t = repr_rec (wakener_repr t) in
+  match t.state with
+    | Sleep{ waiters = waiters } ->
+        let state = Fail e in
+        t.state <- state;
+        run_waiters waiters state
+    | Fail Canceled ->
+        ()
+    | _ ->
+        invalid_arg "Lwt.wakeup_exn"
 
 let restart_cancel t =
   let t = repr_rec (wakener_repr t) in
-  match !t with
+  match t.state with
     | Sleep{ waiters = waiters; cancel = cancel } ->
         let state = Fail Canceled in
-        t := state;
+        t.state <- state;
         run_waiters waiters state
     | _ ->
         ()
 
-let wakeup t v = restart t (Return v) "Lwt.wakeup"
-let wakeup_exn t e = restart t (Fail e) "Lwt.wakeup_exn"
-
 let cancel t =
-  match !(repr t) with
+  match (repr t).state with
     | Sleep{ cancel = cancel } ->
         let f = !cancel in
         cancel := ignore;
-        f ()
+        let save = !current_data in
+        f ();
+        current_data := save
     | _ ->
         ()
 
@@ -172,22 +216,22 @@ let rec cleanup = function
 *)
 let connect t1 t2 =
   let t1 = repr t1 and t2 = repr t2 in
-  match !t1 with
+  match t1.state with
     | Sleep sleeper1 ->
         if t1 == t2 then
           (* Do nothing if the two threads already have the same
              representation *)
           ()
         else begin
-          match !t2 with
+          match t2.state with
             | Sleep sleeper2 ->
                 (* If [t2] is sleeping, then makes it behave as [t1]: *)
-                t2 := Repr t1;
+                t2.state <- Repr t1;
                 (* Note that the order is important: the user have no
                    access to [t2] but may keep a reference to [t1]. If
                    we inverse the order, i.e. we do:
 
-                   [t1 := Repr t2]
+                   [t1.state <- Repr t2]
 
                    then we have a possible leak. For example:
 
@@ -208,9 +252,9 @@ let connect t1 t2 =
                 *)
 
                 (* However, since [t1] is a temporary thread created
-                   for a thread that terminated, its cancel function
-                   is meaningless. Only the one of [t2] is now
-                   important: *)
+                   for a thread that is now terminated, its cancel
+                   function is meaningless. Only the one of [t2] is
+                   now important: *)
                 sleeper1.cancel := !(sleeper2.cancel);
 
                 (* Merge the two sets of waiters: *)
@@ -225,8 +269,8 @@ let connect t1 t2 =
                   sleeper1.waiters <- waiters
                 end
             | state2 ->
-                (* [t2] has already terminated, assing its state to [t1]: *)
-                t1 := state2;
+                (* [t2] is already terminated, assing its state to [t1]: *)
+                t1.state <- state2;
                 (* and run all the waiters of [t1]: *)
                 run_waiters sleeper1.waiters state2
         end
@@ -237,9 +281,9 @@ let connect t1 t2 =
 (* Same as [connect] except that we know that [t2] has terminated: *)
 let fast_connect t state =
   let t = repr t in
-  match !t with
+  match t.state with
     | Sleep{ waiters = waiters } ->
-        t := state;
+        t.state <- state;
         run_waiters waiters state
     | _ ->
         invalid_arg "Lwt.fast_connect"
@@ -249,25 +293,36 @@ let fast_connect t state =
    +-----------------------------------------------------------------+ *)
 
 let return v =
-  thread (ref (Return v))
+  thread { state = Return v }
+
 let fail e =
-  thread (ref (Fail e))
+  thread { state = Fail e }
+
 let temp r =
-  thread (ref (Sleep{ cancel = r;
-                      waiters = Empty;
-                      removed = 0 }))
-let no_cancel = ref ignore
+  thread {
+    state = Sleep{ cancel = r;
+                   waiters = Empty;
+                   removed = 0 }
+  }
+
 let wait () =
-  let t = ref (Sleep{ cancel = no_cancel;
-                      waiters = Empty;
-                      removed = 0 }) in
+  let t = {
+    state = Sleep{ cancel = ref ignore;
+                   waiters = Empty;
+                   removed = 0 };
+  } in
   (thread t, wakener t)
+
 let task () =
-  let rec t = ref (Sleep{ cancel = ref (fun () -> restart_cancel (wakener t));
-                          waiters = Empty;
-                          removed = 0 }) in
+  let rec t = {
+    state = Sleep{ cancel = ref (fun () -> restart_cancel (wakener t));
+                   waiters = Empty;
+                   removed = 0 };
+  } in
   (thread t, wakener t)
 
+let waiter_of_wakener wakener = thread (wakener_repr wakener)
+
 (* apply function, reifying explicit exceptions into the thread type
    apply: ('a -(exn)-> 'b t) -> ('a -(n)-> 'b t)
    semantically a natural transformation TE -> T, where T is the thread
@@ -286,12 +341,12 @@ let add_removable_waiter sleeper waiter =
                         | _ -> Append(Removable waiter, sleeper.waiters))
 
 let on_cancel t f =
-  let t = repr t in
-  match !t with
+  match (repr t).state with
     | Sleep sleeper ->
+        let data = !current_data in
         add_immutable_waiter sleeper
           (function
-             | Fail Canceled -> (try f () with _ -> ())
+             | Fail Canceled -> current_data := data; (try f () with _ -> ())
              | _ -> ())
     | Fail Canceled ->
         f ()
@@ -299,38 +354,54 @@ let on_cancel t f =
         ()
 
 let bind t f =
-  match !(repr t) with
+  match (repr t).state with
     | Return v ->
-        (* we don't use apply here so that tail recursion is not
-           broken *)
         f v
-    | Fail e ->
-        fail e
+    | Fail exn ->
+        fail exn
     | Sleep sleeper ->
         let res = temp sleeper.cancel in
+        let data = !current_data in
         add_immutable_waiter sleeper
           (function
-             | Return v -> connect res (try f v with exn -> fail exn)
+             | Return v -> current_data := data; connect res (try f v with exn -> fail exn)
              | Fail exn -> fast_connect res (Fail exn)
              | _ -> assert false);
         res
     | Repr _ ->
         assert false
 
+let on_success t f =
+  match (repr t).state with
+    | Return v ->
+        f v
+    | Fail exn ->
+        raise exn
+    | Sleep sleeper ->
+        let data = !current_data in
+        add_immutable_waiter sleeper
+          (function
+             | Return v -> current_data := data; f v
+             | Fail exn -> raise exn
+             | _ -> assert false)
+    | Repr _ ->
+        assert false
+
 let (>>=) t f = bind t f
 let (=<<) f t = bind t f
 
 let map f t =
-  match !(repr t) with
+  match (repr t).state with
     | Return v ->
         return (f v)
     | Fail e ->
-        fail e
+        thread { state = Fail e }
     | Sleep sleeper ->
         let res = temp sleeper.cancel in
+        let data = !current_data in
         add_immutable_waiter sleeper
           (function
-             | Return v -> fast_connect res (try Return(f v) with exn -> Fail exn)
+             | Return v -> current_data := data; fast_connect res (try Return(f v) with exn -> Fail exn)
              | Fail exn -> fast_connect res (Fail exn)
              | _ -> assert false);
         res
@@ -342,50 +413,68 @@ let (=|<) f t = map f t
 
 let catch x f =
   let t = try x () with exn -> fail exn in
-  match !(repr t) with
-    | Return v ->
+  match (repr t).state with
+    | Return _ ->
         t
     | Fail exn ->
         f exn
     | Sleep sleeper ->
         let res = temp sleeper.cancel in
+        let data = !current_data in
         add_immutable_waiter sleeper
           (function
              | Return _ as state -> fast_connect res state
-             | Fail exn -> connect res (try f exn with exn -> fail exn)
+             | Fail exn -> current_data := data; connect res (try f exn with exn -> fail exn)
              | _ -> assert false);
         res
     | Repr _ ->
         assert false
 
+let on_failure t f =
+  match (repr t).state with
+    | Return v ->
+        ()
+    | Fail exn ->
+        f exn
+    | Sleep sleeper ->
+        let data = !current_data in
+        add_immutable_waiter sleeper
+          (function
+             | Return v -> ()
+             | Fail exn -> current_data := data; f exn
+             | _ -> assert false)
+    | Repr _ ->
+        assert false
+
 let try_bind x f g =
   let t = try x () with exn -> fail exn in
-  match !(repr t) with
+  match (repr t).state with
     | Return v ->
         f v
     | Fail exn ->
         g exn
     | Sleep sleeper ->
         let res = temp sleeper.cancel in
+        let data = !current_data in
         add_immutable_waiter sleeper
           (function
-             | Return v -> connect res (try f v with exn -> fail exn)
-             | Fail exn -> connect res (try g exn with exn -> fail exn)
+             | Return v -> current_data := data; connect res (try f v with exn -> fail exn)
+             | Fail exn -> current_data := data; connect res (try g exn with exn -> fail exn)
              | _ -> assert false);
         res
     | Repr _ ->
         assert false
 
 let poll t =
-  match !(repr t) with
+  match (repr t).state with
     | Fail e -> raise e
     | Return v -> Some v
     | Sleep _ -> None
     | Repr _ -> assert false
 
 let rec ignore_result t =
-  match !(repr t) with
-    | Return v ->
+  match (repr t).state with
+    | Return _ ->
         ()
     | Fail e ->
         raise e
@@ -399,7 +488,7 @@ let rec ignore_result t =
         assert false
 
 let protected t =
-  match !(repr t) with
+  match (repr t).state with
     | Sleep sleeper ->
         let waiter, wakener = task () in
         add_immutable_waiter sleeper
@@ -421,23 +510,22 @@ let rec nth_ready l n =
   match l with
     | [] ->
         assert false
-    | x :: rem ->
-        let x = repr x in
-        match !x with
+    | t :: l ->
+        match (repr t).state with
           | Sleep _ ->
-              nth_ready rem n
+              nth_ready l n
           | _ when n > 0 ->
-              nth_ready rem (n - 1)
-          | _ ->
-              thread x
+              nth_ready l (n - 1)
+          | state ->
+              state
 
 let ready_count l =
-  List.fold_left (fun acc x -> match !(repr x) with Sleep _ -> acc | _ -> acc + 1) 0 l
+  List.fold_left (fun acc x -> match (repr x).state with Sleep _ -> acc | _ -> acc + 1) 0 l
 
 let remove_waiters l =
   List.iter
     (fun t ->
-       match !(repr t) with
+       match (repr t).state with
          | Sleep sleeper ->
              let removed = sleeper.removed + 1 in
              if removed > max_removed then begin
@@ -452,11 +540,15 @@ let remove_waiters l =
 let choose l =
   let ready = ready_count l in
   if ready > 0 then
-    nth_ready l (Random.int ready)
+    if ready = 1 then
+      (* Optimisation for the common case: *)
+      thread { state = nth_ready l 0 }
+    else
+      thread { state = nth_ready l (Random.int ready) }
   else begin
     let res = temp (ref (fun () -> List.iter cancel l)) in
-    let rec waiter = ref (Some handle_result)
-    and handle_result state =
+    let waiter = ref None in
+    let handle_result state =
       (* Disable the waiter now: *)
       waiter := None;
       (* Removes all waiters so we do not leak memory: *)
@@ -465,10 +557,14 @@ let choose l =
          since all other waiters have been removed: *)
       fast_connect res state
     in
+    waiter := (Some handle_result);
     List.iter
       (fun t ->
-         match !(repr t) with
+         match (repr t).state with
            | Sleep sleeper ->
+               (* The data passed here will never be used because
+                  [handle_result] only calls [fast_connect] which
+                  calls [run_waiters] which ignore the current data *)
                add_removable_waiter sleeper waiter;
            | _ ->
                assert false)
@@ -480,11 +576,11 @@ let rec nchoose_terminate res acc = function
   | [] ->
       fast_connect res (Return(List.rev acc))
   | t :: l ->
-      match !(repr t) with
+      match (repr t).state with
         | Return x ->
             nchoose_terminate res (x :: acc) l
-        | Fail _ as state ->
-            fast_connect res state
+        | Fail e ->
+            fast_connect res (Fail e)
         | _ ->
             nchoose_terminate res acc l
 
@@ -498,7 +594,7 @@ let nchoose_sleep l =
   in
   List.iter
     (fun t ->
-       match !(repr t) with
+       match (repr t).state with
          | Sleep sleeper ->
              add_removable_waiter sleeper waiter;
          | _ ->
@@ -511,8 +607,7 @@ let nchoose l =
     | [] ->
         nchoose_sleep l
     | t :: l ->
-        let t = repr t in
-        match !t with
+        match (repr t).state with
           | Return x ->
               collect [x] l
           | Fail exn ->
@@ -523,8 +618,7 @@ let nchoose l =
     | [] ->
         return (List.rev acc)
     | t :: l ->
-        let t = repr t in
-        match !t with
+        match (repr t).state with
           | Return x ->
               collect (x :: acc) l
           | Fail exn ->
@@ -534,27 +628,86 @@ let nchoose l =
   in
   init l
 
+let rec nchoose_split_terminate res acc_terminated acc_sleeping = function
+  | [] ->
+      fast_connect res (Return(List.rev acc_terminated, List.rev acc_sleeping))
+  | t :: l ->
+      match (repr t).state with
+        | Return x ->
+            nchoose_split_terminate res (x :: acc_terminated) acc_sleeping l
+        | Fail e ->
+            fast_connect res (Fail e)
+        | _ ->
+            nchoose_split_terminate res acc_terminated (t :: acc_sleeping) l
+
+let nchoose_split_sleep l =
+  let res = temp (ref (fun () -> List.iter cancel l)) in
+  let rec waiter = ref (Some handle_result)
+  and handle_result state =
+    waiter := None;
+    remove_waiters l;
+    nchoose_split_terminate res [] [] l
+  in
+  List.iter
+    (fun t ->
+       match (repr t).state with
+         | Sleep sleeper ->
+             add_removable_waiter sleeper waiter;
+         | _ ->
+             assert false)
+    l;
+  res
+
+let nchoose_split l =
+  let rec init acc_sleeping = function
+    | [] ->
+        nchoose_split_sleep l
+    | t :: l ->
+        match (repr t).state with
+          | Return x ->
+              collect [x] acc_sleeping l
+          | Fail exn ->
+              fail exn
+          | _ ->
+              init (t :: acc_sleeping) l
+  and collect acc_terminated acc_sleeping = function
+    | [] ->
+        return (List.rev acc_terminated, acc_sleeping)
+    | t :: l ->
+        match (repr t).state with
+          | Return x ->
+              collect (x :: acc_terminated) acc_sleeping l
+          | Fail exn ->
+              fail exn
+          | _ ->
+              collect acc_terminated (t :: acc_sleeping) l
+  in
+  init [] l
+
 (* Return the nth ready thread, and cancel all others *)
 let rec cancel_and_nth_ready l n =
   match l with
     | [] ->
         assert false
-    | x :: rem ->
-        let x = repr x in
-        match !x with
+    | t :: l ->
+        match (repr t).state with
           | Sleep _ ->
-              cancel (thread x);
-              nth_ready rem n
+              cancel t;
+              nth_ready l n
           | _ when n > 0 ->
-              nth_ready rem (n - 1)
-          | _ ->
-              List.iter cancel rem;
-              thread x
+              nth_ready l (n - 1)
+          | state ->
+              List.iter cancel l;
+              state
 
 let pick l =
   let ready = ready_count l in
   if ready > 0 then
-    cancel_and_nth_ready l (Random.int ready)
+    if ready = 1 then
+      (* Optimisation for the common case: *)
+      thread { state = cancel_and_nth_ready l 0 }
+    else
+      thread { state = cancel_and_nth_ready l (Random.int ready) }
   else begin
     let res = temp (ref (fun () -> List.iter cancel l)) in
     let rec waiter = ref (Some handle_result)
@@ -567,7 +720,7 @@ let pick l =
     in
     List.iter
       (fun t ->
-         match !(repr t) with
+         match (repr t).state with
            | Sleep sleeper ->
                add_removable_waiter sleeper waiter;
            | _ ->
@@ -587,7 +740,7 @@ let npick_sleep l =
   in
   List.iter
     (fun t ->
-       match !(repr t) with
+       match (repr t).state with
          | Sleep sleeper ->
              add_removable_waiter sleeper waiter;
          | _ ->
@@ -600,8 +753,7 @@ let npick threads =
     | [] ->
         npick_sleep threads
     | t :: l ->
-        let t = repr t in
-        match !t with
+        match (repr t).state with
           | Return x ->
               collect [x] l
           | Fail exn ->
@@ -614,8 +766,7 @@ let npick threads =
         List.iter cancel threads;
         return (List.rev acc)
     | t :: l ->
-        let t = repr t in
-        match !t with
+        match (repr t).state with
           | Return x ->
               collect (x :: acc) l
           | Fail exn ->
@@ -627,60 +778,46 @@ let npick threads =
   init threads
 
 let join l =
-  let res = temp (ref (fun () -> List.iter cancel l)) and sleeping = ref 0 (* Number of threads still sleeping *) in
+  let res = temp (ref (fun () -> List.iter cancel l))
+  (* Number of threads still sleeping: *)
+  and sleeping = ref 0
+  (* The state that must be returned: *)
+  and return_state = ref (Return ()) in
   let rec waiter = ref (Some handle_result)
-  and handle_result state = match state with
-    | Fail exn ->
-        (* The thread has failed, exit immediatly without waiting for
-           other threads *)
-        waiter := None;
-        remove_waiters l;
-        fast_connect res state
-    | _ ->
-        decr sleeping;
-        (* Every threads has finished, we can wakeup the result: *)
-        if !sleeping = 0 then begin
-          waiter := None;
-          fast_connect res state
-        end
+  and handle_result state =
+    begin
+      match !return_state, state with
+        | Return _, Fail exn -> return_state := state
+        | _ -> ()
+    end;
+    decr sleeping;
+    (* All threads are terminated, we can wakeup the result: *)
+    if !sleeping = 0 then begin
+      waiter := None;
+      fast_connect res !return_state
+    end
   in
   let rec init = function
     | [] ->
         if !sleeping = 0 then
-          (* No threads is sleeping, returns immediately: *)
-          return ()
+          (* No thread is sleeping, returns immediately: *)
+          thread { state = !return_state }
         else
           res
     | t :: rest ->
-        match !(repr t) with
-          | Fail exn ->
-              (* One of the thread already failed, remove the waiter
-                 from all already visited sleeping threads and
-                 fail: *)
-              let rec loop = function
-                | [] ->
-                    t
-                | t :: l ->
-                    match !(repr t) with
-                      | Fail _ ->
-                          t
-                      | Sleep sleeper ->
-                          let removed = sleeper.removed + 1 in
-                          if removed > max_removed then begin
-                            sleeper.removed <- 0;
-                            sleeper.waiters <- cleanup sleeper.waiters
-                          end else
-                            sleeper.removed <- removed;
-                          loop l
-                      | _ ->
-                          loop l
-              in
-              waiter := None;
-              loop l
+        match (repr t).state with
           | Sleep sleeper ->
               incr sleeping;
               add_removable_waiter sleeper waiter;
               init rest
+          | Fail _ as state -> begin
+              match !return_state with
+                | Return _ ->
+                    return_state := state;
+                    init rest
+                | _ ->
+                    init rest
+            end
           | _ ->
               init rest
   in
@@ -694,6 +831,128 @@ let finalize f g =
     (fun x -> g () >>= fun () -> return x)
     (fun e -> g () >>= fun () -> fail e)
 
+let update_data key = function
+  | Some _ as value ->
+      current_data := Int_map.add key.id (fun () -> key.store <- value) !current_data
+  | None ->
+      current_data := Int_map.remove key.id !current_data
+
+let with_value key value f =
+  let save = !current_data in
+  let data =
+    match value with
+      | Some _ ->
+          Int_map.add key.id (fun () -> key.store <- value) save
+      | None ->
+          Int_map.remove key.id save
+  in
+  current_data := data;
+  try
+    let result = f () in
+    current_data := save;
+    result
+  with exn ->
+    current_data := save;
+    raise exn
+
+(* +-----------------------------------------------------------------+
+   | Paused threads                                                  |
+   +-----------------------------------------------------------------+ *)
+
+let pause_hook = ref ignore
+
+let paused = Lwt_sequence.create ()
+let paused_count = ref 0
+
+let pause () =
+  let waiter, wakener = task () in
+  let node = Lwt_sequence.add_r wakener paused in
+  on_cancel waiter (fun () -> Lwt_sequence.remove node);
+  incr paused_count;
+  !pause_hook !paused_count;
+  waiter
+
+let wakeup_paused () =
+  if not (Lwt_sequence.is_empty paused) then begin
+    let tmp = Lwt_sequence.create () in
+    Lwt_sequence.transfer_r paused tmp;
+    paused_count := 0;
+    Lwt_sequence.iter_l (fun wakener -> wakeup wakener ()) tmp
+  end
+
+let register_pause_notifier f = pause_hook := f
+
+let paused_count () = !paused_count
+
+(* +-----------------------------------------------------------------+
+   | Bakctrace support                                               |
+   +-----------------------------------------------------------------+ *)
+
+let backtrace_bind add_loc t f =
+  match (repr t).state with
+    | Return v ->
+        f v
+    | Fail exn ->
+        thread { state = Fail(add_loc exn) }
+    | Sleep sleeper ->
+        let res = temp sleeper.cancel in
+        let data = !current_data in
+        add_immutable_waiter sleeper
+          (function
+             | Return v -> current_data := data; connect res (try f v with exn -> fail (add_loc exn))
+             | Fail exn -> fast_connect res (Fail(add_loc exn))
+             | _ -> assert false);
+        res
+    | Repr _ ->
+        assert false
+
+let backtrace_catch add_loc x f =
+  let t = try x () with exn -> fail exn in
+  match (repr t).state with
+    | Return _ ->
+        t
+    | Fail exn ->
+        f (add_loc exn)
+    | Sleep sleeper ->
+        let res = temp sleeper.cancel in
+        let data = !current_data in
+        add_immutable_waiter sleeper
+          (function
+             | Return _ as state -> fast_connect res state
+             | Fail exn -> current_data := data; connect res (try f exn with exn -> fail (add_loc exn))
+             | _ -> assert false);
+        res
+    | Repr _ ->
+        assert false
+
+let backtrace_try_bind add_loc x f g =
+  let t = try x () with exn -> fail exn in
+  match (repr t).state with
+    | Return v ->
+        f v
+    | Fail exn ->
+        g (add_loc exn)
+    | Sleep sleeper ->
+        let res = temp sleeper.cancel in
+        let data = !current_data in
+        add_immutable_waiter sleeper
+          (function
+             | Return v -> current_data := data; connect res (try f v with exn -> fail (add_loc exn))
+             | Fail exn -> current_data := data; connect res (try g exn with exn -> fail (add_loc exn))
+             | _ -> assert false);
+        res
+    | Repr _ ->
+        assert false
+
+let backtrace_finalize add_loc f g =
+  backtrace_try_bind add_loc f
+    (fun x -> g () >>= fun () -> return x)
+    (fun e -> g () >>= fun () -> fail (add_loc e))
+
+(* +-----------------------------------------------------------------+
+   | Threads state query                                             |
+   +-----------------------------------------------------------------+ *)
+
 module State = struct
   type 'a state =
     | Return of 'a
@@ -701,26 +960,10 @@ module State = struct
     | Sleep
 end
 
-let state t = match !(repr t) with
+let state t = match (repr t).state with
   | Return v -> State.Return v
   | Fail exn -> State.Fail exn
   | Sleep _ -> State.Sleep
   | Repr _ -> assert false
 
 include State
-
-let paused = Lwt_sequence.create ()
-
-let pause () =
-  let waiter, wakener = task () in
-  let node = Lwt_sequence.add_r wakener paused in
-  on_cancel waiter (fun () -> Lwt_sequence.remove node);
-  waiter
-
-let rec wakeup_paused () =
-  match Lwt_sequence.take_opt_l paused with
-    | None ->
-        ()
-    | Some wakener ->
-        wakeup wakener ();
-        wakeup_paused ()
