@@ -37,13 +37,6 @@ let check_buffer_size fun_name buffer_size =
 
 let default_buffer_size = ref 4096
 
-let close_fd fd =
-  try
-    Lwt_unix.close fd;
-    return ()
-  with exn ->
-    fail exn
-
 (* +-----------------------------------------------------------------+
    | Types                                                           |
    +-----------------------------------------------------------------+ *)
@@ -87,7 +80,7 @@ and 'mode channel = {
 }
 
 and 'mode _channel = {
-  mutable buffer : string;
+  mutable buffer : Lwt_bytes.t;
   mutable length : int;
 
   mutable ptr : int;
@@ -122,18 +115,18 @@ and 'mode _channel = {
 }
 
 and typ =
-  | Type_normal of (string -> int -> int -> int Lwt.t) * (int64 -> Unix.seek_command -> int64 Lwt.t)
+  | Type_normal of (Lwt_bytes.t -> int -> int -> int Lwt.t) * (int64 -> Unix.seek_command -> int64 Lwt.t)
       (* The channel has been created with [make]. The first argument
          is the refill/flush function and the second is the seek
          function. *)
-  | Type_string
-      (* The channel has been created with [of_string]. *)
+  | Type_bytes
+      (* The channel has been created with [of_bytes]. *)
 
 type input_channel = input channel
 type output_channel = output channel
 
 type direct_access = {
-  da_buffer : string;
+  da_buffer : Lwt_bytes.t;
   mutable da_ptr : int;
   mutable da_max : int;
   da_perform : unit -> int Lwt.t;
@@ -170,6 +163,15 @@ let name ch = match ch.mode with
 let closed_channel ch = Channel_closed(name ch)
 let invalid_channel ch = Failure(Printf.sprintf "temporary atomic %s channel no more valid" (name ch))
 
+let is_busy ch =
+  match ch.state with
+    | Invalid ->
+        raise (invalid_channel ch.channel)
+    | Idle | Closed ->
+        false
+    | Busy_primitive | Busy_atomic _ ->
+        true
+
 (* Flush/refill the buffer. No race condition could happen because
    this function is always called atomically: *)
 let perform_io ch = match ch.main.state with
@@ -181,7 +183,7 @@ let perform_io ch = match ch.main.state with
                   (* Size of data in the buffer *)
                   let size = ch.max - ch.ptr in
                   (* If there are still data in the buffer, keep them: *)
-                  if size > 0 then String.unsafe_blit ch.buffer ch.ptr ch.buffer 0 size;
+                  if size > 0 then Lwt_bytes.unsafe_blit ch.buffer ch.ptr ch.buffer 0 size;
                   (* Update positions: *)
                   ch.ptr <- 0;
                   ch.max <- size;
@@ -191,8 +193,8 @@ let perform_io ch = match ch.main.state with
             lwt n = pick [ch.abort_waiter; perform_io ch.buffer ptr len] in
             (* Never trust user functions... *)
             if n < 0 || n > len then
-              fail (Failure (Printf.sprintf "Lwt_io: invalid result of the [%s] function(request=%d,result=%d)"
-                               (match ch.mode with Input -> "read" | Output -> "write") len n))
+              raise_lwt (Failure (Printf.sprintf "Lwt_io: invalid result of the [%s] function(request=%d,result=%d)"
+                                    (match ch.mode with Input -> "read" | Output -> "write") len n))
             else begin
               (* Update the global offset: *)
               ch.offset <- Int64.add ch.offset (Int64.of_int n);
@@ -203,26 +205,26 @@ let perform_io ch = match ch.main.state with
                 | Output ->
                     (* Shift remaining data: *)
                     let len = len - n in
-                    String.unsafe_blit ch.buffer n ch.buffer 0 len;
+                    Lwt_bytes.unsafe_blit ch.buffer n ch.buffer 0 len;
                     ch.ptr <- len
               end;
               return n
             end
 
-        | Type_string -> begin
+        | Type_bytes -> begin
             match ch.mode with
               | Input ->
                   return 0
               | Output ->
-                  fail (Failure "cannot flush a channel created with Lwt_io.of_string")
+                  raise_lwt (Failure "cannot flush a channel created with Lwt_io.of_string")
           end
     end
 
   | Closed ->
-      fail (closed_channel ch)
+      raise_lwt (closed_channel ch)
 
   | Invalid ->
-      fail (invalid_channel ch)
+      raise_lwt (invalid_channel ch)
 
   | Idle ->
       assert false
@@ -330,7 +332,7 @@ let primitive f wrapper = match wrapper.state with
         | Closed ->
             (* The channel has been closed while we were waiting *)
             unlock wrapper;
-            fail (closed_channel wrapper.channel)
+            raise_lwt (closed_channel wrapper.channel)
 
         | Idle ->
             wrapper.state <- Busy_primitive;
@@ -341,17 +343,17 @@ let primitive f wrapper = match wrapper.state with
               return ()
 
         | Invalid ->
-            fail (invalid_channel wrapper.channel)
+            raise_lwt (invalid_channel wrapper.channel)
 
         | Busy_primitive | Busy_atomic _ ->
             assert false
       end
 
   | Closed ->
-      fail (closed_channel wrapper.channel)
+      raise_lwt (closed_channel wrapper.channel)
 
   | Invalid ->
-      fail (invalid_channel wrapper.channel)
+      raise_lwt (invalid_channel wrapper.channel)
 
 (* Wrap a sequence of io operations into an atomic operation: *)
 let atomic f wrapper = match wrapper.state with
@@ -377,7 +379,7 @@ let atomic f wrapper = match wrapper.state with
         | Closed ->
             (* The channel has been closed while we were waiting *)
             unlock wrapper;
-            fail (closed_channel wrapper.channel)
+            raise_lwt (closed_channel wrapper.channel)
 
         | Idle ->
             let tmp_wrapper = { state = Idle;
@@ -392,17 +394,17 @@ let atomic f wrapper = match wrapper.state with
               return ()
 
         | Invalid ->
-            fail (invalid_channel wrapper.channel)
+            raise_lwt (invalid_channel wrapper.channel)
 
         | Busy_primitive | Busy_atomic _ ->
             assert false
       end
 
   | Closed ->
-      fail (closed_channel wrapper.channel)
+      raise_lwt (closed_channel wrapper.channel)
 
   | Invalid ->
-      fail (invalid_channel wrapper.channel)
+      raise_lwt (invalid_channel wrapper.channel)
 
 let rec abort wrapper = match wrapper.state with
   | Busy_atomic tmp_wrapper ->
@@ -412,7 +414,7 @@ let rec abort wrapper = match wrapper.state with
       (* Double close, just returns the same thing as before *)
       Lazy.force wrapper.channel.close
   | Invalid ->
-      fail (invalid_channel wrapper.channel)
+      raise_lwt (invalid_channel wrapper.channel)
   | Idle | Busy_primitive ->
       wrapper.state <- Closed;
       (* Abort any current real reading/writing operation on the
@@ -423,7 +425,7 @@ let rec abort wrapper = match wrapper.state with
 let close wrapper =
   let channel = wrapper.channel in
   if channel.main != wrapper then
-    fail (Failure "Lwt_io.close: cannot close a channel obtained via Lwt_io.atomic")
+    raise_lwt (Failure "Lwt_io.close: cannot close a channel obtained via Lwt_io.atomic")
   else
     match channel.mode with
       | Input ->
@@ -451,26 +453,27 @@ let () =
          wrappers)
 
 let no_seek pos cmd =
-  fail (Failure "Lwt_io.seek: seek not supported on this channel")
+  raise_lwt (Failure "Lwt_io.seek: seek not supported on this channel")
 
 external unsafe_output : 'a channel -> output channel = "%identity"
 
 let make ?buffer_size ?(close=return) ?(seek=no_seek) ~mode perform_io =
-  let buffer =
-    String.create (match buffer_size with
-                     | None ->
-                         !default_buffer_size
-                     | Some size ->
-                         check_buffer_size "Lwt_io.make" size;
-                         size)
-  and abort_waiter, abort_wakener = Lwt.wait () in
+  let size =
+    match buffer_size with
+      | None ->
+          !default_buffer_size
+      | Some size ->
+          check_buffer_size "Lwt_io.make" size;
+          size
+  in
+  let buffer = Lwt_bytes.create size and abort_waiter, abort_wakener = Lwt.wait () in
   let rec ch = {
     buffer = buffer;
-    length = String.length buffer;
+    length = size;
     ptr = 0;
     max = (match mode with
              | Input -> 0
-             | Output -> String.length buffer);
+             | Output -> size);
     close = lazy(try_lwt close ());
     abort_waiter = abort_waiter;
     abort_wakener = abort_wakener;
@@ -478,7 +481,7 @@ let make ?buffer_size ?(close=return) ?(seek=no_seek) ~mode perform_io =
     auto_flushing = false;
     mode = mode;
     offset = 0L;
-    typ = Type_normal(perform_io, fun pos cmd -> try seek pos cmd with e -> fail e);
+    typ = Type_normal(perform_io, fun pos cmd -> try seek pos cmd with e -> raise_lwt e);
   } and wrapper = {
     state = Idle;
     channel = ch;
@@ -487,11 +490,11 @@ let make ?buffer_size ?(close=return) ?(seek=no_seek) ~mode perform_io =
   if mode = Output then Outputs.add outputs (unsafe_output wrapper);
   wrapper
 
-let of_string ~mode str =
-  let length = String.length str in
+let of_bytes ~mode bytes =
+  let length = Lwt_bytes.length bytes in
   let abort_waiter, abort_wakener = Lwt.wait () in
   let rec ch = {
-    buffer = str;
+    buffer = bytes;
     length = length;
     ptr = 0;
     max = length;
@@ -504,7 +507,7 @@ let of_string ~mode str =
     auto_flushing = true;
     mode = mode;
     offset = 0L;
-    typ = Type_string;
+    typ = Type_bytes;
   } and wrapper = {
     state = Idle;
     channel = ch;
@@ -512,27 +515,21 @@ let of_string ~mode str =
   } in
   wrapper
 
+let of_string ~mode str = of_bytes ~mode (Lwt_bytes.of_string str)
+
 let of_fd ?buffer_size ?close ~mode fd =
   let perform_io = match mode with
-    | Input -> Lwt_unix.read fd
-    | Output -> Lwt_unix.write fd
+    | Input -> Lwt_bytes.read fd
+    | Output -> Lwt_bytes.write fd
   in
   make
     ?buffer_size
     ~close:(match close with
               | Some f -> f
-              | None -> (fun () -> close_fd fd))
-    ~seek:(fun pos cmd -> return (Lwt_unix.LargeFile.lseek fd pos cmd))
+              | None -> (fun () -> Lwt_unix.close fd))
+    ~seek:(fun pos cmd -> Lwt_unix.LargeFile.lseek fd pos cmd)
     ~mode
-    (match (Lwt_unix.fstat fd).Unix.st_kind with
-       | Unix.S_REG | Unix.S_DIR ->
-           (* Non-blocking I/Os are not possible on regular files: *)
-           let auto_yield = Lwt_unix.auto_yield 0.05 in
-           (fun str pos ofs ->
-              lwt () = auto_yield () in
-              perform_io str pos ofs)
-       | _ ->
-           perform_io)
+    perform_io
 
 let of_unix_fd ?buffer_size ?close ~mode fd =
   of_fd ?buffer_size ?close ~mode (Lwt_unix.of_unix_file_descr fd)
@@ -547,8 +544,8 @@ let buffer_size ch = ch.channel.length
 let resize_buffer wrapper len =
   if len < min_buffer_size then invalid_arg "Lwt_io.resize_buffer";
   match wrapper.channel.typ with
-    | Type_string ->
-        fail (Failure "Lwt_io.resize_buffer: cannot resize the buffer of a channel created with Lwt_io.of_string")
+    | Type_bytes ->
+        raise_lwt (Failure "Lwt_io.resize_buffer: cannot resize the buffer of a channel created with Lwt_io.of_string")
     | Type_normal _ ->
         primitive begin fun ch ->
           match ch.mode with
@@ -557,10 +554,10 @@ let resize_buffer wrapper len =
                 (* Fail if we want to decrease the buffer size and there is
                    too much unread data in the buffer: *)
                 if len < unread_count then
-                  fail (Failure "Lwt_io.resize_buffer: cannot decrease buffer size")
+                  raise_lwt (Failure "Lwt_io.resize_buffer: cannot decrease buffer size")
                 else begin
-                  let buffer = String.create len in
-                  String.unsafe_blit ch.buffer ch.ptr buffer 0 unread_count;
+                  let buffer = Lwt_bytes.create len in
+                  Lwt_bytes.unsafe_blit ch.buffer ch.ptr buffer 0 unread_count;
                   ch.buffer <- buffer;
                   ch.length <- len;
                   ch.ptr <- 0;
@@ -578,8 +575,8 @@ let resize_buffer wrapper len =
                     return ()
                 in
                 lwt () = loop () in
-                let buffer = String.create len in
-                String.unsafe_blit ch.buffer 0 buffer 0 ch.ptr;
+                let buffer = Lwt_bytes.create len in
+                Lwt_bytes.unsafe_blit ch.buffer 0 buffer 0 ch.ptr;
                 ch.buffer <- buffer;
                 ch.length <- len;
                 ch.max <- len;
@@ -661,11 +658,11 @@ struct
     let ptr = ic.ptr in
     if ptr = ic.max then
       refill ic >>= function
-        | 0 -> fail End_of_file
+        | 0 -> raise_lwt End_of_file
         | _ -> read_char ic
     else begin
       ic.ptr <- ptr + 1;
-      return (String.unsafe_get ic.buffer ptr)
+      return (Lwt_bytes.unsafe_get ic.buffer ptr)
     end
 
   let read_char_opt ic =
@@ -693,7 +690,7 @@ struct
                if cr_read then Buffer.add_char buf '\r';
                return(Buffer.contents buf)
            | exn ->
-               fail exn)
+               raise_lwt exn)
     in
     read_char ic >>= function
       | '\r' -> loop true
@@ -710,13 +707,13 @@ struct
     let avail = ic.max - ic.ptr in
     if avail > 0 then begin
       let len = min len avail in
-      String.unsafe_blit ic.buffer ic.ptr str ofs len;
+      Lwt_bytes.unsafe_blit_bytes_string ic.buffer ic.ptr str ofs len;
       ic.ptr <- ic.ptr + len;
       return len
     end else begin
       refill ic >>= fun n ->
         let len = min len n in
-        String.unsafe_blit ic.buffer 0 str ofs len;
+        Lwt_bytes.unsafe_blit_bytes_string ic.buffer 0 str ofs len;
         ic.ptr <- len;
         ic.max <- n;
         return len
@@ -724,9 +721,9 @@ struct
 
   let read_into ic str ofs len =
     if ofs < 0 || len < 0 || ofs + len > String.length str then
-      fail (Invalid_argument (Printf.sprintf
-                                "Lwt_io.read_into(ofs=%d,len=%d,str_len=%d)"
-                                ofs len (String.length str)))
+      raise_lwt (Invalid_argument (Printf.sprintf
+                                     "Lwt_io.read_into(ofs=%d,len=%d,str_len=%d)"
+                                     ofs len (String.length str)))
     else begin
       if len = 0 then
         return 0
@@ -737,7 +734,7 @@ struct
   let rec unsafe_read_into_exactly ic str ofs len =
     unsafe_read_into ic str ofs len >>= function
       | 0 ->
-          fail End_of_file
+          raise_lwt End_of_file
       | n ->
           let len = len - n in
           if len = 0 then
@@ -747,9 +744,9 @@ struct
 
   let read_into_exactly ic str ofs len =
     if ofs < 0 || len < 0 || ofs + len > String.length str then
-      fail (Invalid_argument (Printf.sprintf
-                                "Lwt_io.read_into_exactly(ofs=%d,len=%d,str_len=%d)"
-                                ofs len (String.length str)))
+      raise_lwt (Invalid_argument (Printf.sprintf
+                                     "Lwt_io.read_into_exactly(ofs=%d,len=%d,str_len=%d)"
+                                     ofs len (String.length str)))
     else begin
       if len = 0 then
         return ()
@@ -757,19 +754,34 @@ struct
         unsafe_read_into_exactly ic str ofs len
     end
 
-  let rec read_all ic buf =
-    Buffer.add_substring buf ic.buffer ic.ptr (ic.max - ic.ptr);
+  let rev_concat len l =
+    let buf = String.create len in
+    let _ =
+      List.fold_left
+        (fun ofs str ->
+           let len = String.length str in
+           let ofs = ofs - len in
+           String.unsafe_blit str 0 buf ofs len;
+           ofs)
+        len l
+    in
+    buf
+
+  let rec read_all ic total_len acc =
+    let len = ic.max - ic.ptr in
+    let str = String.create len in
+    Lwt_bytes.unsafe_blit_bytes_string ic.buffer ic.ptr str 0 len;
     ic.ptr <- ic.max;
     refill ic >>= function
       | 0 ->
-          return (Buffer.contents buf)
+          return (rev_concat (len + total_len) (str :: acc))
       | n ->
-          read_all ic buf
+          read_all ic (len + total_len) (str :: acc)
 
   let read count ic =
     match count with
       | None ->
-          read_all ic (Buffer.create 512)
+          read_all ic 0 []
       | Some len ->
           let str = String.create len in
           lwt real_len = unsafe_read_into ic str 0 len in
@@ -783,7 +795,7 @@ struct
     lwt () = unsafe_read_into_exactly ic header 0 20 in
     let bsize = Marshal.data_size header 0 in
     let buffer = String.create (20 + bsize) in
-    String.unsafe_blit header 0 buffer 0 20 ;
+    String.unsafe_blit header 0 buffer 0 20;
     lwt () = unsafe_read_into_exactly ic buffer 20 bsize in
     return (Marshal.from_string buffer 0)
 
@@ -797,7 +809,7 @@ struct
     let ptr = oc.ptr in
     if ptr < oc.length then begin
       oc.ptr <- ptr + 1;
-      String.unsafe_set oc.buffer ptr ch;
+      Lwt_bytes.unsafe_set oc.buffer ptr ch;
       return ()
     end else
       lwt _ = flush_partial oc in
@@ -806,11 +818,11 @@ struct
   let rec unsafe_write_from oc str ofs len =
     let avail = oc.length - oc.ptr in
     if avail >= len then begin
-      String.unsafe_blit str ofs oc.buffer oc.ptr len;
+      Lwt_bytes.unsafe_blit_string_bytes str ofs oc.buffer oc.ptr len;
       oc.ptr <- oc.ptr + len;
       return 0
     end else begin
-      String.unsafe_blit str ofs oc.buffer oc.ptr avail;
+      Lwt_bytes.unsafe_blit_string_bytes str ofs oc.buffer oc.ptr avail;
       oc.ptr <- oc.length;
       lwt _ = flush_partial oc in
       let len = len - avail in
@@ -828,9 +840,9 @@ struct
 
   let write_from oc str ofs len =
     if ofs < 0 || len < 0 || ofs + len > String.length str then
-      fail (Invalid_argument (Printf.sprintf
-                                "Lwt_io.write_from(ofs=%d,len=%d,str_len=%d)"
-                                ofs len (String.length str)))
+      raise_lwt (Invalid_argument (Printf.sprintf
+                                     "Lwt_io.write_from(ofs=%d,len=%d,str_len=%d)"
+                                     ofs len (String.length str)))
     else begin
       if len = 0 then
         return 0
@@ -847,9 +859,9 @@ struct
 
   let write_from_exactly oc str ofs len =
     if ofs < 0 || len < 0 || ofs + len > String.length str then
-      fail (Invalid_argument (Printf.sprintf
-                                "Lwt_io.write_from_exactly(ofs=%d,len=%d,str_len=%d)"
-                                ofs len (String.length str)))
+      raise_lwt (Invalid_argument (Printf.sprintf
+                                     "Lwt_io.write_from_exactly(ofs=%d,len=%d,str_len=%d)"
+                                     ofs len (String.length str)))
     else begin
       if len = 0 then
         return ()
@@ -875,7 +887,7 @@ struct
     if ic.max - ic.ptr < size then
       refill ic >>= function
         | 0 ->
-            fail End_of_file
+            raise_lwt End_of_file
         | _ ->
             read_block_unsafe ic size f
     else begin
@@ -896,7 +908,7 @@ struct
 
   let block ch size f =
     if size < 0 || size > min_buffer_size then
-      fail (Invalid_argument(Printf.sprintf "Lwt_io.block(size=%d)" size))
+      raise_lwt (Invalid_argument(Printf.sprintf "Lwt_io.block(size=%d)" size))
     else
       if ch.max - ch.ptr >= size then begin
         let ptr = ch.ptr in
@@ -912,7 +924,7 @@ struct
   let perform token da ch =
     if !token then begin
       if da.da_max <> ch.max || da.da_ptr < ch.ptr || da.da_ptr > ch.max then
-        fail (Invalid_argument "Lwt_io.direct_access.perform")
+        raise_lwt (Invalid_argument "Lwt_io.direct_access.perform")
       else begin
         ch.ptr <- da.da_ptr;
         lwt count = perform_io ch in
@@ -921,7 +933,7 @@ struct
         return count
       end
     end else
-      fail (Failure "Lwt_io.direct_access.perform: this function can not be called outside Lwt_io.direct_access")
+      raise_lwt (Failure "Lwt_io.direct_access.perform: this function can not be called outside Lwt_io.direct_access")
 
   let direct_access ch f =
     let token = ref true in
@@ -934,7 +946,7 @@ struct
     lwt x = f da in
     token := false;
     if da.da_max <> ch.max || da.da_ptr < ch.ptr || da.da_ptr > ch.max then
-      fail (Failure "Lwt_io.direct_access: invalid result of [f]")
+      raise_lwt (Failure "Lwt_io.direct_access: invalid result of [f]")
     else begin
       ch.ptr <- da.da_ptr;
       return x
@@ -948,7 +960,7 @@ struct
        | Reading numbers                                             |
        +-------------------------------------------------------------+ *)
 
-    let get buffer ptr = Char.code (String.unsafe_get buffer ptr)
+    let get buffer ptr = Char.code (Lwt_bytes.unsafe_get buffer ptr)
 
     let read_int ic =
       read_block_unsafe ic 4
@@ -1023,7 +1035,7 @@ struct
        | Writing numbers                                             |
        +-------------------------------------------------------------+ *)
 
-    let set buffer ptr x = String.unsafe_set buffer ptr (Char.unsafe_chr x)
+    let set buffer ptr x = Lwt_bytes.unsafe_set buffer ptr (Char.unsafe_chr x)
 
     let write_int oc v =
       write_block_unsafe oc 4
@@ -1074,7 +1086,7 @@ struct
   let do_seek seek pos =
     lwt offset = seek pos Unix.SEEK_SET in
     if offset <> pos then
-      fail (Failure "Lwt_io.set_position: seek failed")
+      raise_lwt (Failure "Lwt_io.set_position: seek failed")
     else
       return ()
 
@@ -1096,9 +1108,9 @@ struct
           ch.max <- 0;
           return ()
         end
-    | Type_string, _ ->
+    | Type_bytes, _ ->
         if pos < 0L || pos > Int64.of_int ch.length then
-          fail (Failure "Lwt_io.set_position: out of bounds")
+          raise_lwt (Failure "Lwt_io.set_position: out of bounds")
         else begin
           ch.ptr <- Int64.to_int pos;
           return ()
@@ -1109,7 +1121,7 @@ struct
         lwt len = seek 0L Unix.SEEK_END in
         lwt () = do_seek seek ch.offset in
         return len
-    | Type_string ->
+    | Type_bytes ->
         return (Int64.of_int ch.length)
 end
 
@@ -1124,7 +1136,7 @@ let read_char wrapper =
      increases performances by 10x. *)
   if wrapper.state = Idle && ptr < channel.max then begin
     channel.ptr <- ptr + 1;
-    return (String.unsafe_get channel.buffer ptr)
+    return (Lwt_bytes.unsafe_get channel.buffer ptr)
   end else
     primitive Primitives.read_char wrapper
 
@@ -1133,7 +1145,7 @@ let read_char_opt wrapper =
   let ptr = channel.ptr in
   if wrapper.state = Idle && ptr < channel.max then begin
     channel.ptr <- ptr + 1;
-    return (Some(String.unsafe_get channel.buffer ptr))
+    return (Some(Lwt_bytes.unsafe_get channel.buffer ptr))
   end else
     primitive Primitives.read_char_opt wrapper
 
@@ -1151,7 +1163,7 @@ let write_char wrapper x =
   let ptr = channel.ptr in
   if wrapper.state = Idle && ptr < channel.max then begin
     channel.ptr <- ptr + 1;
-    String.unsafe_set channel.buffer ptr x;
+    Lwt_bytes.unsafe_set channel.buffer ptr x;
     (* Fast launching of the auto flusher: *)
     if not channel.auto_flushing then begin
       channel.auto_flushing <- true;
@@ -1211,11 +1223,12 @@ end
 module LE = MakeNumberIO(ByteOrder.LE)
 module BE = MakeNumberIO(ByteOrder.BE)
 
-type byte_order = Little_endian | Big_endian
+type byte_order = Lwt_sys.byte_order = Little_endian | Big_endian
+let system_byte_order = Lwt_sys.byte_order
 
-external get_system_byte_order : unit -> byte_order = "lwt_unix_system_byte_order"
-
-let system_byte_order = get_system_byte_order ()
+include (val (match system_byte_order with
+                | Little_endian -> (module LE : NumberIO)
+                | Big_endian -> (module BE : NumberIO)) : NumberIO)
 
 (* +-----------------------------------------------------------------+
    | Other                                                           |
@@ -1230,7 +1243,7 @@ let zero =
   make
     ~mode:input
     ~buffer_size:min_buffer_size
-    (fun str ofs len -> String.fill str ofs len '\x00'; return len)
+    (fun str ofs len -> Lwt_bytes.fill str ofs len '\x00'; return len)
 
 let null =
   make
@@ -1281,10 +1294,11 @@ let open_file ?buffer_size ?flags ?perm ~mode filename =
     | None, Output ->
         0o666
   in
-  of_unix_fd ?buffer_size ~mode (Unix.openfile filename flags perm)
+  lwt fd = Lwt_unix.openfile filename flags perm in
+  return (of_fd ?buffer_size ~mode fd)
 
 let with_file ?buffer_size ?flags ?perm ~mode filename f =
-  let ic = open_file ?buffer_size ?flags ?perm ~mode filename in
+  lwt ic = open_file ?buffer_size ?flags ?perm ~mode filename in
   try_lwt
     f ic
   finally
@@ -1295,21 +1309,27 @@ let file_length filename = with_file ~mode:input filename length
 let open_connection ?buffer_size sockaddr =
   let fd = Lwt_unix.socket (Unix.domain_of_sockaddr sockaddr) Unix.SOCK_STREAM 0 in
   let close = lazy begin
-    Lwt_unix.shutdown fd Unix.SHUTDOWN_ALL;
-    close_fd fd
+    try_lwt
+      Lwt_unix.shutdown fd Unix.SHUTDOWN_ALL;
+      return ()
+    with Unix.Unix_error(Unix.ENOTCONN, _, _) ->
+      (* This may happen if the server closed the connection before us *)
+      return ()
+    finally
+      Lwt_unix.close fd
   end in
   try_lwt
     lwt () = Lwt_unix.connect fd sockaddr in
     (try Lwt_unix.set_close_on_exec fd with Invalid_argument _ -> ());
     return (make ?buffer_size
               ~close:(fun _ -> Lazy.force close)
-              ~mode:input (Lwt_unix.read fd),
+              ~mode:input (Lwt_bytes.read fd),
             make ?buffer_size
               ~close:(fun _ -> Lazy.force close)
-              ~mode:output (Lwt_unix.write fd))
+              ~mode:output (Lwt_bytes.write fd))
   with exn ->
-    lwt () = close_fd fd in
-    fail exn
+    lwt () = Lwt_unix.close fd in
+    raise_lwt exn
 
 let with_connection ?buffer_size sockaddr f =
   lwt ic, oc = open_connection sockaddr in
@@ -1337,13 +1357,13 @@ let establish_server ?buffer_size ?(backlog=5) sockaddr f =
           (try Lwt_unix.set_close_on_exec fd with Invalid_argument _ -> ());
           let close = lazy begin
             Lwt_unix.shutdown fd Unix.SHUTDOWN_ALL;
-            close_fd fd
+            Lwt_unix.close fd
           end in
           f (of_fd ?buffer_size ~mode:input ~close:(fun () -> Lazy.force close) fd,
              of_fd ?buffer_size ~mode:output ~close:(fun () -> Lazy.force close) fd);
           loop ()
       | `Shutdown ->
-          Lwt_unix.close sock;
+          lwt () = Lwt_unix.close sock in
           match sockaddr with
             | Unix.ADDR_UNIX path when path <> "" && path.[0] <> '\x00' ->
                 Unix.unlink path;
@@ -1357,9 +1377,14 @@ let establish_server ?buffer_size ?(backlog=5) sockaddr f =
 let ignore_close ch =
   ignore (close ch)
 
-let make_stream f ic =
-  Gc.finalise ignore_close ic;
+let make_stream f lazy_ic =
+  let lazy_ic =
+    lazy(lwt ic = Lazy.force lazy_ic in
+         Gc.finalise ignore_close ic;
+         return ic)
+  in
   Lwt_stream.from (fun _ ->
+                     lwt ic = Lazy.force lazy_ic in
                      lwt x = f ic in
                      if x = None then
                        lwt () = close ic in
@@ -1368,13 +1393,13 @@ let make_stream f ic =
                        return x)
 
 let lines_of_file filename =
-  make_stream read_line_opt (open_file ~mode:input filename)
+  make_stream read_line_opt (lazy(open_file ~mode:input filename))
 
 let lines_to_file filename lines =
   with_file ~mode:output filename (fun oc -> write_lines oc lines)
 
 let chars_of_file filename =
-  make_stream read_char_opt (open_file ~mode:input filename)
+  make_stream read_char_opt (lazy(open_file ~mode:input filename))
 
 let chars_to_file filename chars =
   with_file ~mode:output filename (fun oc -> write_chars oc chars)
