@@ -52,7 +52,7 @@ let string_of_level = function
   | Fatal -> "fatal"
 
 (* +-----------------------------------------------------------------+
-   | Patterns andrules                                               |
+   | Patterns and rules                                              |
    +-----------------------------------------------------------------+ *)
 
 type pattern = string list
@@ -106,7 +106,7 @@ let split pattern =
   in
   loop 0
 
-let rules =
+let rules = ref (
   match try Some(Sys.getenv "LWT_LOG") with Not_found -> None with
     | Some str ->
         let rec loop = function
@@ -126,6 +126,7 @@ let rules =
         loop (Lwt_log_rules.rules (Lexing.from_string str))
     | None ->
         []
+)
 
 (* +-----------------------------------------------------------------+
    | Sections                                                        |
@@ -136,19 +137,46 @@ struct
   type t = {
     name : string;
     mutable level : level;
+    mutable modified : bool;
   }
 
-  let make name =
-    let rec find_level = function
+  type section = t
+
+  module Sections = Weak.Make(struct
+                                type t = section
+                                let equal a b = a.name = b.name
+                                let hash s = Hashtbl.hash s.name
+                              end)
+
+  let sections = Sections.create 32
+
+  let find_level name =
+    let rec loop = function
       | [] ->
           Notice
       | (pattern, level) :: rest ->
           if pattern_match pattern name then
             level
           else
-            find_level rest
+            loop rest
     in
-    { name = name; level = find_level rules }
+    loop !rules
+
+  let recompute_levels () =
+    Sections.iter
+      (fun section ->
+         if not section.modified then
+           section.level <- find_level section.name)
+      sections
+
+  let make name =
+    let section = { name = name; level = Notice; modified = false } in
+    try
+      Sections.find sections section
+    with Not_found ->
+      section.level <- find_level name;
+      Sections.add sections section;
+      section
 
   let name section = section.name
 
@@ -156,11 +184,27 @@ struct
 
   let level section = section.level
 
-  let set_level section level = section.level <- level
+  let set_level section level =
+    section.level <- level;
+    section.modified <- true
+
+  let reset_level section =
+    if section.modified then begin
+      section.modified <- false;
+      section.level <- find_level section.name
+    end
 end
 
 type section = Section.t
 
+let add_rule pattern level =
+  rules := (split pattern, level) :: !rules;
+  Section.recompute_levels ()
+
+let append_rule pattern level =
+  rules := !rules @ [(split pattern, level)];
+  Section.recompute_levels ()
+
 (* +-----------------------------------------------------------------+
    | Loggers                                                         |
    +-----------------------------------------------------------------+ *)
@@ -201,8 +245,8 @@ let dispatch f =
 
 type template = string
 
-let date_string () =
-  let tm = Unix.localtime (Unix.time ()) in
+let date_string time =
+  let tm = Unix.localtime time in
   let month_string =
     match tm.Unix.tm_mon with
       | 0 -> "Jan"
@@ -222,9 +266,11 @@ let date_string () =
   Printf.sprintf "%s %2d %02d:%02d:%02d" month_string tm.Unix.tm_mday tm.Unix.tm_hour tm.Unix.tm_min tm.Unix.tm_sec
 
 let render ~buffer ~template ~section ~level ~message =
+  let time = lazy(Unix.gettimeofday ()) in
   Buffer.add_substitute buffer
     (function
-       | "date" -> date_string ()
+       | "date" -> date_string (Lazy.force time)
+       | "milliseconds" -> String.sub (Printf.sprintf "%.4f" (fst (modf (Lazy.force time)))) 2 4
        | "name" -> program_name
        | "pid" -> string_of_int (Unix.getpid ())
        | "message" -> message
@@ -271,10 +317,10 @@ let file ?(template="$(date): $(section): $(message)") ?(mode=`Append) ?(perm=0o
         [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_APPEND; Unix.O_NONBLOCK]
     | `Truncate ->
         [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC; Unix.O_NONBLOCK] in
-  let fd = Lwt_unix.openfile file_name flags 0o666 in
+  lwt fd = Lwt_unix.openfile file_name flags 0o666 in
   Lwt_unix.set_close_on_exec fd;
   let oc = Lwt_io.of_fd ~mode:Lwt_io.output fd in
-  channel ~template ~close_mode:`Close ~channel:oc ()
+  return (channel ~template ~close_mode:`Close ~channel:oc ())
 
 let level_code = function
   | Fatal -> 0
@@ -329,7 +375,7 @@ let syslog_connect paths =
         (* No working socket found *)
         log_intern "no working socket found in {%s}; is syslogd running?"
           (String.concat ", " (List.map (Printf.sprintf "\"%s\"") paths));
-        fail (Sys_error(Unix.error_message Unix.ENOENT))
+        raise_lwt (Sys_error(Unix.error_message Unix.ENOENT))
     | path :: paths ->
         begin try
           return (Some (Unix.stat path).Unix.st_kind)
@@ -351,7 +397,7 @@ let syslog_connect paths =
                 return (DGRAM, fd)
               with
                 | Unix.Unix_error(Unix.EPROTOTYPE, _, _) -> begin
-                    Lwt_unix.close fd;
+                    lwt () = Lwt_unix.close fd in
                     (* Then try with a stream socket: *)
                     let fd = Lwt_unix.socket Unix.PF_UNIX Unix.SOCK_STREAM 0 in
                     try_lwt
@@ -359,12 +405,12 @@ let syslog_connect paths =
                       Lwt_unix.set_close_on_exec fd;
                       return (STREAM, fd)
                     with Unix.Unix_error(error, _, _) ->
-                      Lwt_unix.close fd;
+                      lwt () = Lwt_unix.close fd in
                       log_intern "can not connect to \"%s\": %s" path (Unix.error_message error);
                       loop paths
                   end
                 | Unix.Unix_error(error, _, _) ->
-                    Lwt_unix.close fd;
+                    lwt () = Lwt_unix.close fd in
                     log_intern "can not connect to \"%s\": %s" path (Unix.error_message error);
                     loop paths
             end
@@ -431,7 +477,7 @@ let syslog ?(template="$(date) $(name)[$(pid)]: $(section): $(message)") ?(paths
                             print socket_type fd lines
                           with Unix.Unix_error(_, _, _) ->
                             (* Try to reconnect *)
-                            shutdown fd;
+                            lwt () = shutdown fd in
                             syslog_socket := None;
                             lwt socket_type, fd = get_syslog () in
                             lwt () = write_string fd (make_line socket_type line) in
@@ -444,8 +490,7 @@ let syslog ?(template="$(date) $(name)[$(pid)]: $(section): $(message)") ?(paths
                 | None ->
                     return ()
                 | Some(socket_type, fd) ->
-                    shutdown fd;
-                    return ())
+                    shutdown fd)
 
 (* +-----------------------------------------------------------------+
    | Logging functions                                               |
@@ -468,7 +513,7 @@ let log ?exn ?(section=Section.main) ?logger ~level message =
     | Some logger -> logger
   in
   if logger.lg_closed then
-    fail Logger_closed
+    raise_lwt Logger_closed
   else if level >= section.Section.level then
     match exn with
       | None ->
